#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.interfaces module

This module provides all package interfaces.
"""

from zope.interface import Interface, implementer
from zope.interface.interfaces import IObjectEvent, ObjectEvent


DISABLE_CATALOG_SETTINGS_KEY = 'pyams_catalog.disable_indexing'

NLTK_LANGUAGES = {
    'da': 'danish',
    'nl': 'dutch',
    'en': 'english',
    'fi': 'finnish',
    'fr': 'french',
    'de': 'german',
    'hu': 'hungarian',
    'it': 'italian',
    'no': 'norwegian',
    'po': 'porter',
    'pt': 'portuguese',
    'ro': 'romanian',
    'ru': 'russian',
    'es': 'spanish',
    'sv': 'swedish'
}

NO_RESOLUTION = 6
SECOND_RESOLUTION = 5
MINUTE_RESOLUTION = 4
HOUR_RESOLUTION = 3
DATE_RESOLUTION = 2
MONTH_RESOLUTION = 1
YEAR_RESOLUTION = 0


class INoAutoIndex(Interface):
    """Marker interface for objects which shouldn't be automatically indexed"""


class IBeforeObjectIndexEvent(IObjectEvent):
    """Interface of event fired before object indexation

    This event can be useful to be able to initialize some attributes before object
    is registered into the catalog.
    """


@implementer(IBeforeObjectIndexEvent)
class BeforeObjectIndexEvent(ObjectEvent):
    """Event fired before object indexation"""



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.generations main module

This module is checking for registered utilities on site upgrade, and provides functions which
can be used by generations utilities to check for missing indexes.
"""

import logging

from hypatia.catalog import Catalog
from hypatia.interfaces import ICatalog
from zope.interface.interface import InterfaceClass
from zope.intid import IIntIds, IntIds

from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'

LOGGER = logging.getLogger('PyAMS (catalog)')

RENAMED_CLASSES = {
    'pyams_i18n.index I18nTextIndexMixin': 'pyams_catalog.i18n I18nTextIndexMixin',
    'pyams_i18n.index I18nTextIndexWithInterface': 'pyams_catalog.i18n I18nTextIndexWithInterface'
}

REQUIRED_UTILITIES = ((IIntIds, '', IntIds, 'Internal IDs'),
                      (ICatalog, '', Catalog, 'Catalog'),)


@utility_config(name='PyAMS catalog', provides=ISiteGenerations)
class CatalogGenerationsChecker:
    """Catalog generations checker"""

    order = 30
    generation = 1

    def evolve(self, site, current=None):  # pylint: disable=no-self-use,unused-argument
        """Check for required utilities"""
        check_required_utilities(site, REQUIRED_UTILITIES)


def check_required_indexes(site, indexes, catalog_name=''):
    """Utility function to check for required catalog indexes

    utilities argument is a tuple made of:
    - the index name
    - the index class
    - index factory arguments
    """
    sm = site.getSiteManager()  # pylint: disable=invalid-name
    catalog = sm.queryUtility(ICatalog, name=catalog_name)
    if catalog is None:
        LOGGER.warning("No catalog found! Index check ignored...")
        return
    for name, klass, args in indexes:
        index = catalog.get(name)
        if index is None:
            for key, value in args.copy().items():
                if callable(value) and not isinstance(value, InterfaceClass):
                    args[key] = value()
            index = klass(**args)
            catalog[name] = index



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.i18n module
"""

from ZODB.broken import Broken
from hypatia.text import TextIndex
from hypatia.text.lexicon import Lexicon
from hypatia.util import BaseIndexMixin
from persistent import Persistent

from pyams_catalog.nltk import NltkFullTextProcessor
from pyams_i18n.interfaces import II18n


__docformat__ = 'restructuredtext'


_MARKER = object()


class I18nTextIndexMixin(BaseIndexMixin):
    """I18n text index mixin"""

    def __init__(self, language, interface=None):
        self.interface = interface
        self.language = language

    def discriminate(self, obj, default):
        if self.interface is not None:
            obj = self.interface(obj, None)
            if obj is None:
                return default

        # pylint: disable=assignment-from-no-return
        value = II18n(obj).get_attribute(self.discriminator, lang=self.language, default=_MARKER)
        if value is _MARKER:
            return default

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object {0!r}'.format(value))

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object {0!r}'.format(value))

        return value


class I18nTextIndexWithInterface(I18nTextIndexMixin, TextIndex):
    """I18n text index"""

    # pylint: disable=too-many-arguments
    def __init__(self, language, discriminator, interface=None, lexicon=None, index=None,
                 family=None):
        I18nTextIndexMixin.__init__(self, language, interface)
        if lexicon is None:
            lexicon = Lexicon(NltkFullTextProcessor(language))
        TextIndex.__init__(self, discriminator, lexicon, index, family)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.utils module

This module defines several utility functions which are used to update catalog.
"""

import logging

import transaction
from ZODB.interfaces import IBroken
from hypatia.interfaces import ICatalog
from zope.interface import Interface
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import NotYet

from pyams_catalog.interfaces import BeforeObjectIndexEvent, INoAutoIndex
from pyams_site.site import site_factory
from pyams_utils.adapter import adapter_config
from pyams_utils.container import find_objects_providing
from pyams_utils.interfaces import ICacheKeyValue
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility, \
    set_local_registry


__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (catalog)')


@adapter_config(required=ICatalog, provides=ICacheKeyValue)
def catalog_key_adapter(obj):
    """Catalog key value adapter"""
    return 'catalog::{}'.format(str(sorted(obj)))


def index_object(obj, catalog='', ignore_notyet=False):
    """Index given object into catalog"""
    LOGGER.debug("Indexing object {0!r}".format(obj))
    intids = query_utility(IIntIds)
    if intids is not None:
        try:
            object_id = intids.register(obj)
        except NotYet:
            if not ignore_notyet:
                raise
        else:
            if isinstance(catalog, str):
                catalog = query_utility(ICatalog, name=catalog)
            if catalog is not None:
                get_pyramid_registry().notify(BeforeObjectIndexEvent(obj))
                catalog.index_doc(object_id, obj)


def reindex_object(obj, catalog=''):
    """Reindex given object into catalog"""
    LOGGER.debug("Re-indexing object {0!r}".format(obj))
    intids = query_utility(IIntIds)
    if intids is not None:
        object_id = intids.queryId(obj)
        if object_id is not None:
            if isinstance(catalog, str):
                catalog = query_utility(ICatalog, name=catalog)
            if catalog is not None:
                get_pyramid_registry().notify(BeforeObjectIndexEvent(obj))
                catalog.reindex_doc(object_id, obj)


def unindex_object(obj, catalog=''):
    """Un-index given object from catalog"""
    LOGGER.debug("Un-indexing object {0!r}".format(obj))
    intids = query_utility(IIntIds)
    if intids is not None:
        object_id = intids.queryId(obj)
        if object_id is not None:
            if isinstance(catalog, str):
                catalog = query_utility(ICatalog, name=catalog)
            if catalog is not None:
                catalog.unindex_doc(object_id)


def index_site(request, autocommit=True):
    """Index all site contents in internal catalog"""
    application = site_factory(request)
    if application is not None:
        try:
            set_local_registry(application.getSiteManager())
            catalog = get_utility(ICatalog)
            catalog.reset()
            transaction.savepoint()
            intids = get_utility(IIntIds)
            for index, document in enumerate(find_objects_providing(application, Interface)):
                if INoAutoIndex.providedBy(document):
                    continue
                if IBroken.providedBy(document):
                    print("Skipping broken object: {0!r}".format(document))
                else:
                    print("Indexing: {0!r}".format(document))
                    catalog.reindex_doc(intids.register(document), document)
                    if not index % 100:
                        transaction.savepoint()
        finally:
            set_local_registry(None)
        if autocommit:
            transaction.commit()
    return application



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""
Generic test case for pyams_catalog docstrings
"""

__docformat__ = 'restructuredtext'

import doctest
import os
import unittest

from pyams_catalog.tests import get_package_dir


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def doc_suite(test_dir, globs=None):
    """Returns a test suite, based on doc tests strings found in /*.py"""
    suite = []
    if globs is None:
        globs = globals()

    flags = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE |
             doctest.REPORT_ONLY_FIRST_FAILURE)

    package_dir = get_package_dir(test_dir)

    # filtering files on extension
    docs = [doc for doc in
            os.listdir(package_dir) if doc.endswith('.py')]
    docs = [doc for doc in docs if not doc.startswith('__')]

    for test in docs:
        fd = open(os.path.join(package_dir, test))  # pylint: disable=invalid-name
        content = fd.read()
        fd.close()
        if '>>> ' not in content:
            continue
        test = test.replace('.py', '')
        location = 'pyams_catalog.%s' % test
        suite.append(doctest.DocTestSuite(location, optionflags=flags,
                                          globs=globs))

    return unittest.TestSuite(suite)


def test_suite():
    """returns the test suite"""
    return doc_suite(CURRENT_DIR)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.events module

This module register events subscribers which are automatically used to index, reindex or
unindex a content during it's lifecycle when it is added, modified or removed.
"""

from hypatia.interfaces import ICatalog
from persistent import IPersistent
from pyramid.settings import asbool
from pyramid.events import subscriber
from zope.intid.interfaces import IIntIdRemovedEvent
from zope.lifecycleevent import IObjectAddedEvent, IObjectModifiedEvent

from pyams_catalog.interfaces import DISABLE_CATALOG_SETTINGS_KEY
from pyams_catalog.utils import index_object, reindex_object, unindex_object
from pyams_utils.registry import get_utilities_for, get_pyramid_registry


__docformat__ = 'restructuredtext'


@subscriber(IObjectAddedEvent, context_selector=IPersistent)
def handle_new_object(event):
    """Index new persistent object"""
    registry = get_pyramid_registry()
    if asbool(registry.settings.get(DISABLE_CATALOG_SETTINGS_KEY)):
        return
    for _, catalog in get_utilities_for(ICatalog):
        index_object(event.object, catalog, ignore_notyet=True)


@subscriber(IObjectModifiedEvent, context_selector=IPersistent)
def handle_modified_object(event):
    """Update catalog for modified persistent object"""
    registry = get_pyramid_registry()
    if asbool(registry.settings.get(DISABLE_CATALOG_SETTINGS_KEY)):
        return
    for _, catalog in get_utilities_for(ICatalog):
        reindex_object(event.object, catalog)


@subscriber(IIntIdRemovedEvent, context_selector=IPersistent)
def handle_removed_object(event):
    """Un-index removed persistent object

    Don't use IObjectRemovedEvent to avoid objects from being already unregistered
    from IIntId utility!!!
    """
    registry = get_pyramid_registry()
    if asbool(registry.settings.get(DISABLE_CATALOG_SETTINGS_KEY)):
        return
    for _, catalog in get_utilities_for(ICatalog):
        unindex_object(event.object, catalog)
